/** @file camogm_jpeg.c
 * @brief Provides writing to series of individual JPEG files for @e camogm
 * @copyright Copyright (C) 2016 Elphel, Inc.
 *
 * @par <b>License</b>
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/** @brief This define is needed to use lseek64 and should be set before includes */
#define _LARGEFILE64_SOURCE
//#define __USE_GNU // for O_DIRECT
#define _GNU_SOURCE // for O_DIRECT
#include <linux/raw.h> // for O_DIRECT
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
//#include <bits/fcntl.h> // 'O_DIRECT'
#include <sys/uio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <time.h>
#include <elphel/c313a.h>
#include <elphel/ahci_cmd.h>

#include "camogm_jpeg.h"
#include "camogm_read.h"
#include "camogm_align.h"
//#define BUFFER_KVEC
#define MAX_IMAGE_ALIGNED (4096*256)
/** State file record format. It includes device path in /dev, starting, current and ending LBAs */
#define STATE_FILE_FORMAT         "%s\t%llu\t%llu\t%llu\n"
/** Status file update period, in seconds */
#define STAT_UPDATE_PERIOD        60

/* forward declarations */
static void *jpeg_writer(void *thread_args);
static int save_state_file(const camogm_state *state);

/**
 * Emulate writev() by consecutive write()
 * @param __fd
 * @param __iovec
 * @param __count
 * @return
 */
ssize_t emul_writev (int fd, const struct iovec *iovec, int count) {
	int i;
	ssize_t wlen, tot_len = 0;
	for (i = 0; i < count; i++) {
		if (iovec[i].iov_len > 0) {
			if (iovec[i].iov_len & (PAGE_PHYS -1)){
				D0(fprintf(debug_file, "Length not page-aligned: while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
						iovec[i].iov_len, iovec[i].iov_base, ((uint64_t*) iovec[i].iov_base)[0],i ));
				D0(syslog (LOG_INFO, "Length not page-aligned: while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
						iovec[i].iov_len, iovec[i].iov_base, ((uint64_t*) iovec[i].iov_base)[0],i));
				debug_level = 6;
				return -1;
			}
			wlen = write (fd, iovec[i].iov_base, iovec[i].iov_len);
			if (wlen < 0) {
				D0(fprintf(debug_file, "Error %d while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
						wlen, iovec[i].iov_len, iovec[i].iov_base, ((uint64_t*) iovec[i].iov_base)[0],i ));
				return wlen;
			} else if (wlen < iovec[i].iov_len) {
				return tot_len + wlen;
			} else {
				tot_len += wlen;
			}
		}
	}
	return tot_len;
}

/** Get starting and endign LBAs of the partition specified as raw device buffer */
static int get_disk_range(struct range *range)
{
	FILE *f;

	// get raw device buffer starting postion on disk
	f = fopen(SYSFS_AHCI_LBA_START, "r");
	if (f == NULL) {
		return -1;
	}
	fscanf(f, "%llu\n", &range->from);
	fclose(f);

	// get raw device buffer ending postion on disk
	f = fopen(SYSFS_AHCI_LBA_END, "r");
	if (f == NULL) {
		return -1;
	}
	fscanf(f, "%llu\n", &range->to);
	fclose(f);

	return 0;
}

/** Get write pointer from a file. This functions check not only the name of a partition, but
 * its geometry as well */
static int find_state(FILE *f, uint64_t *pos, const rawdev_buffer *rawdev)
{
	size_t len;
	uint64_t start_pos, curr_pos, end_pos;
	struct range range;
	char buff[ELPHEL_PATH_MAX];
	char dev_name[ELPHEL_PATH_MAX];

	if (f == NULL || pos == NULL)
		return -1;
	if (get_disk_range(&range) != 0) {
		return -1;
	}

	// skip first line containing file header
	fgets(buff, ELPHEL_PATH_MAX, f);
	while (fgets(buff, ELPHEL_PATH_MAX, f) != NULL) {
		sscanf(buff, STATE_FILE_FORMAT, dev_name, &start_pos, &curr_pos, &end_pos);
		len = strlen(dev_name);
		if (strncmp(rawdev->rawdev_path, dev_name, len) == 0 &&
				range.from == start_pos &&
				range.to == end_pos) {
			*pos = curr_pos;
			break;
		}
	}

	return 0;
}

/** Read state from file and restore disk write pointer */
int open_state_file(const rawdev_buffer *rawdev, uint64_t *current_pos)
{
	FILE *f;
	int ret = 0;
	uint64_t lba_pos;

	if (strlen(rawdev->state_path) == 0) {
		return ret;
	}

	f = fopen(rawdev->state_path, "r");
	if (f != NULL) {
		if (find_state(f, &lba_pos, rawdev) != -1) {
			*current_pos = lba_pos;
			D0(fprintf(debug_file, "Got starting LBA from state file: %llu\n", lba_pos));
		}
		fclose(f);
	} else {
		ret = -1;
	}

	return ret;
}

/** Save current position of the disk write pointer */
static int save_state_file(const camogm_state *state)
{
	int ret = 0;
	FILE *f;
	const rawdev_buffer *rawdev = &state->rawdev;
	const struct writer_params *params = &state->writer_params;

	if (strlen(rawdev->state_path) == 0) {
		return ret;
	}

	// save pointers to a regular file
	f = fopen(rawdev->state_path, "w");
	if (f == NULL) {
		return -1;
	}
	fprintf(f, "Device\t\tStart LBA\tCurrent LBA\tEnd LBA\n");
	fprintf(f, STATE_FILE_FORMAT, rawdev->rawdev_path, params->lba_start, params->lba_current, params->lba_end);
	fflush(f);
	fsync(fileno(f));
	fclose(f);

	return ret;
}

/**
 * @brief Initialize synchronization resources for disk writing thread and then start this thread. This function
 * is call each time JPEG format is set or changed, thus we need to check the state of writing thread before
 * initialization to prevent spawning multiple threads.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      0 if initialization was successful and negative value otherwise
 */
int camogm_init_jpeg(camogm_state *state)
{
	int ret = 0;
	int ret_val;

	if (state->writer_params.state == STATE_STOPPED) {
		ret_val = pthread_cond_init(&state->writer_params.main_cond, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize conditional variable for main thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_cond_init(&state->writer_params.writer_cond, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize conditional variable for writing thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_mutex_init(&state->writer_params.writer_mutex, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize mutex for writing thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_create(&state->writer_params.writer_thread, NULL, jpeg_writer, (void *)state);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not start writer thread: %s\n", strerror(ret_val)));
			ret = -1;
		}

		ret_val = init_align_buffers(state);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize alignment buffers\n"));
			ret = -1;
			return ret;
		}
	}

	return ret;
}

/**
 * @brief Stop disk writing thread and free its resources. This function is called after the program receives 'exit' command.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      None
 */
void camogm_free_jpeg(camogm_state *state)
{
	pthread_cond_destroy(&state->writer_params.main_cond);
	pthread_cond_destroy(&state->writer_params.writer_cond);
	pthread_mutex_destroy(&state->writer_params.writer_mutex);

	// terminate writing thread
	pthread_mutex_lock(&state->writer_params.writer_mutex);
	state->writer_params.exit_thread = true;
	pthread_cond_signal(&state->writer_params.writer_cond);
	pthread_mutex_unlock(&state->writer_params.writer_mutex);
	pthread_join(state->writer_params.writer_thread, NULL);
	state->writer_params.exit_thread = false;

	deinit_align_buffers(state);
}

/** Calculate the total length of current frame */
int64_t camogm_get_jpeg_size(camogm_state *state)
{
	int64_t len = 0;

	for (int i = 0; i < state->chunk_index - 1; i++) {
		len += state->packetchunks[i + 1].bytes;
	}

	return len;
}

/**
 * @brief Called every time the JPEG files recording is started.
 *
 * This function checks if the raw device write is initiated and tries to open the device specified. The device
 * will be closed in #camogm_end_jpeg function.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      0 if the device was opened successfully and negative error code otherwise
 */
int camogm_start_jpeg(camogm_state *state)
{
	char * slash;
	int rslt;
	uint64_t offset;

	if (!state->rawdev_op) {
		strcpy(state->path, state->path_prefix);   // make state->path a directory name (will be replaced when the frames will be written)
		slash = strrchr(state->path, '/');
		D2(fprintf(debug_file, "camogm_start_jpeg\n"));
		if (slash) {
			D3(fprintf(debug_file, "Full path %s\n", state->path));
			slash[0] = '\0';                       // truncate path to the directory name
			D3(fprintf(debug_file, "directory path %s\n", state->path));
			rslt = mkdir(state->path, 0777);
			D3(fprintf(debug_file, "mkdir (%s, 0777) returned %d, errno=%d\n", state->path, rslt, errno));
			if ((rslt < 0) && (errno != EEXIST)) { // already exists is OK
				D0(fprintf(debug_file, "Error creating directory %s, errno=%d\n", state->path, errno));
				return -CAMOGM_FRAME_FILE_ERR;
			}
		}
	} else {
		if (open_state_file(&state->rawdev, &state->writer_params.lba_current) != 0) {
			D0(fprintf(debug_file, "Could not get write pointer from state file, recording will start from the beginning of partition: "
					"%s\n", state->rawdev.rawdev_path));
		}
//		state->writer_params.blockdev_fd = open(state->rawdev.rawdev_path, O_RDWR | 0x40000); // __O_DIRECT);  // __O_DIRECT);
		state->writer_params.blockdev_fd = open(state->rawdev.rawdev_path, O_RDWR | O_DIRECT); // __O_DIRECT);  // __O_DIRECT);
		// int mode = fcntl(fd, F_GETFL);
		D0(syslog (LOG_INFO, "Raw %s is opened in 0x%08x mode, __O_DIRECT=0x%08x, O_DIRECT=0x%08x", state->rawdev.rawdev_path, fcntl(state->writer_params.blockdev_fd, F_GETFL), __O_DIRECT, O_DIRECT));
		if (state->writer_params.blockdev_fd < 0) {
			D0(fprintf(debug_file, "Error opening block device: %s\n", state->rawdev.rawdev_path));
			return -CAMOGM_FRAME_FILE_ERR;
		}
		offset = lba_to_offset(state->writer_params.lba_current - state->writer_params.lba_start);
		lseek64(state->writer_params.blockdev_fd, offset, SEEK_SET);
		D3(fprintf(debug_file, "Open block device: %s, offset in bytes: %llu\n", state->rawdev.rawdev_path, offset));
		state->writer_params.stat_update = time(NULL);
	}

	return 0;
}

/**
 * @brief Write single JPEG frame
 *
 * This function will write single JPEG file
 * @param   state   a pointer to a structure containing current state
 * @return
 */
int camogm_frame_jpeg(camogm_state *state)
{
	int i, j;
	ssize_t iovlen, l = 0;
	struct iovec chunks_iovec[8];
	int port = state->port_num;
	time_t curr_time;
	int is_tiff = state->this_frame_params[port].color == COLORMODE_RAW;
    if (is_tiff) {
    	sprintf(state->path, "%s%d_%010ld_%06ld.tiff", state->path_prefix, port, state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec);
    } else {
    	sprintf(state->path, "%s%d_%010ld_%06ld.jpeg", state->path_prefix, port, state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec);
    }
    if (!state->rawdev_op) {
    	l = 0;
    	for (i = 0; i < (state->chunk_index) - 1; i++) {
    		chunks_iovec[i].iov_base = state->packetchunks[i + 1].chunk;
    		chunks_iovec[i].iov_len = state->packetchunks[i + 1].bytes;
    		l += chunks_iovec[i].iov_len;
    	}
    	if (((state->ivf = open(state->path, O_RDWR | O_CREAT, 0777))) < 0) {
    		D0(fprintf(debug_file, "Error opening %s for writing, returned %d, errno=%d\n", state->path, state->ivf, errno));
    		return -CAMOGM_FRAME_FILE_ERR;
    	}
    	iovlen = emul_writev(state->ivf, chunks_iovec, (state->chunk_index) - 1); // FIXME! Remove gaps?
    	if (iovlen < l) {
    		j = errno;
    		D0(fprintf(debug_file, "emul_writev error %d (returned %d, expected %d)\n", j, iovlen, l));
    		close(state->ivf);
    		return -CAMOGM_FRAME_FILE_ERR;
    	}
    	state->rawdev.last_jpeg_size = l;
    	close(state->ivf);
    } else {
    	D6(fprintf(debug_file, "ccam_dma_buf[%d] = %p\n",port, get_ccam_dma_buf(port)));
    	D3(fprintf(debug_file, "_12.7_ camogm_frame_jpeg(): %ld.%06ld:%d @ %07d\n",state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec, port, get_fpga_usec(state->fd_fparmsall[0], 0)));
    	D6(fprintf(debug_file, "\ndump iovect array for port %u, chunk_index=%u, chunk_data_index=%u, data_segments=%u\n", state->port_num, state->chunk_index, state->chunk_data_index, state->data_segments));
    	for (int i = 0; i < state->chunk_index; i++) {
    		D6(fprintf(debug_file, "ptr: %p, length: %ld\n", state->packetchunks[i].chunk, state->packetchunks[i].bytes));
    	}
    	// next frame is ready for recording, signal this to the writer thread
    	D6(fprintf(debug_file, "_13_: Yielding thread before getting a lock (%d, %d):%d @ %07d\n", state->writer_params.chunk_page_prep, state->writer_params.chunk_page_write, state->writer_params.writev_run, get_fpga_usec(state->fd_fparmsall[0], 0)));
    	pthread_yield();
    	pthread_mutex_lock(&state->writer_params.writer_mutex);
    	D6(fprintf(debug_file, "_13a_: waiting for empty buffer (%d, %d):%d @ %07d\n", state->writer_params.chunk_page_prep, state->writer_params.chunk_page_write, state->writer_params.writev_run, get_fpga_usec(state->fd_fparmsall[0], 0)));
    	// do not run ahead by >= FILE_CHUNKS_PAGES (0 or 1 for just FILE_CHUNKS_PAGES == 2)
    	// either buffer is completely empty (state->writer_params->chunk_page_prep == state->writer_params->chunk_page_write)
    	// or it is not full and writev_run is active
    	while (((state->writer_params.chunk_page_prep - state->writer_params.chunk_page_write + (2 * SEGMENTS_PAGES))
    			% (2 * SEGMENTS_PAGES) >= SEGMENTS_PAGES) ||
    			(!state->writer_params.writev_run && (state->writer_params.chunk_page_prep != state->writer_params.chunk_page_write))) {
    		D6(fprintf(debug_file, "_13b_: waiting for empty buffer (%d, %d):%d @ %07d\n", state->writer_params.chunk_page_prep, state->writer_params.chunk_page_write, state->writer_params.writev_run, get_fpga_usec(state->fd_fparmsall[0], 0)));
    		pthread_cond_wait(&state->writer_params.main_cond, &state->writer_params.writer_mutex);
    	}
    	pthread_mutex_unlock(&state->writer_params.writer_mutex);
    	D6(fprintf(debug_file, "_13c_: got empty buffer         (%d, %d):%d @ %07d\n", state->writer_params.chunk_page_prep, state->writer_params.chunk_page_write, state->writer_params.writev_run, get_fpga_usec(state->fd_fparmsall[0], 0)));
    	remap_vectors(state); // now can be used instead of align_frame(state), returns total length
//    	align_frame(state);
    	// proceed if last frame was recorded without errors
    	D6(fprintf(debug_file, "_13d_:frame aligned @ %07d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
    	pthread_mutex_lock(&state->writer_params.writer_mutex);
    	if (state->writer_params.last_ret_val == 0) {
    		state->writer_params.chunk_page_prep = (state->writer_params.chunk_page_prep + 1) % (2 * SEGMENTS_PAGES);
    		//			state->writer_params.data_ready = true;
    		pthread_cond_signal(&state->writer_params.writer_cond);
    		D6(fprintf(debug_file, "_13e_: chunk_page_prep = %d (%d), signal writer_cond @ %07d\n",state->writer_params.chunk_page_prep, state->writer_params.chunk_page_write, get_fpga_usec(state->fd_fparmsall[0], 0)));
    	}
    	pthread_mutex_unlock(&state->writer_params.writer_mutex);
    	if (state->writer_params.last_ret_val != 0) {
    		return state->writer_params.last_ret_val;
    	}
    	// update status file if time has come
    	curr_time = time(NULL);
    	if (difftime(curr_time, state->writer_params.stat_update) > STAT_UPDATE_PERIOD) {
    		save_state_file(state);
    		D6(fprintf(debug_file, "_13f_: saved state file @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
    	}
    }

    return 0;
}

/**
 * @brief Finish JPEG file write operation
 *
 * This function checks whether raw device write was on and closes raw device file.
 * @param   state   a pointer to a structure containing current state
 * @return  0 if the device was closed successfully and -1 otherwise
 */
int camogm_end_jpeg(camogm_state *state) // TODO: Resolve multippage !
{
	int ret = 0;
	int bytes, ceil_size;
	ssize_t iovlen;

	if (state->rawdev_op) {
		// write any remaining data, do not use writer thread as there can be only one block left CHUNK_REM buffer
		pthread_mutex_lock(&state->writer_params.writer_mutex);

		while (state->writer_params.chunk_page_prep != state->writer_params.chunk_page_write)
			// wait for previous frame to be recorded first if it has not be recorded by the moment
			pthread_cond_wait(&state->writer_params.main_cond, &state->writer_params.writer_mutex);
		bytes = state->writer_params.glue_carry_vec.iov_len;
		if (bytes > 0) {
			ceil_size = vectaligntail(&state->writer_params.glue_carry_vec);
			D1(fprintf(debug_file, "Write last block of data, size = %d, full size = %d\n", bytes, ceil_size));
			// the remaining data block is placed in CHUNK_COMMON buffer, write just this buffer
			iovlen = emul_writev(state->writer_params.blockdev_fd,
					&state->writer_params.glue_carry_vec,
					1);
			if (iovlen < bytes) {
				D0(fprintf(debug_file, "emul_writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, bytes));
				state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
			} else {
				// update statistic, just one block written
				state->writer_params.lba_current += 1;
				state->rawdev.total_rec_len += bytes;
			}
			//			reset_chunks(state->writer_params.data_chunks, 1);
			//*************** See if another buffer(s) must be cleared too? or even use chunk_page_write
			reset_segments(state, 1, state->writer_params.chunk_page_prep % SEGMENTS_PAGES);
		}
		pthread_mutex_unlock(&state->writer_params.writer_mutex);

		D6(fprintf(debug_file, "Closing block device %s\n", state->rawdev.rawdev_path));
		ret = close(state->writer_params.blockdev_fd);
		if (ret == -1)
			D0(fprintf(debug_file, "Error: %s\n", strerror(errno)));
		save_state_file(state);
	}
	return ret;
}

/**
 * @brief Disk writing thread. This thread holds local copy of a structure containing current state
 * of the program and updates it on a signal from main thread every time new frame is ready for recording.
 * @param[in]   thread_args   a pointer to a structure containing current state
 * @return      None
 */
void *jpeg_writer(void *thread_args)
{
	struct iovec write_segments[SEGMENTS_NUMBER];// 3
	int num_write_segments, seg_len;
	int dbg_us1=0, dbg_us2=0, dbg_us; // save usec for difference
	long dbg_pos;
	int reset_rem;
	int i, i0, i1;
	ssize_t iovlen, l, ll, l0, l1, blocks_write, wpage;
	bool process = true, has_data_to_write = false;
	struct iovec split_chunk;

	camogm_state *state = (camogm_state *)thread_args;
	struct writer_params *params = &state->writer_params;
	params->state = STATE_RUNNING;
	D6(fprintf(debug_file, "_w000_:  Before jpeg_writer loop @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0))); // before the debug file is set !
	while (process) {
		pthread_mutex_lock(&params->writer_mutex);  // first time before the debug file is set !
		D6(fprintf(debug_file, "_w00_: (%d, %d) @ %07d\n", params->chunk_page_prep, params->chunk_page_write, get_fpga_usec(state->fd_fparmsall[0], 0)));
		while ((params->chunk_page_prep == params->chunk_page_write) && !params->exit_thread) {
			pthread_cond_wait(&params->writer_cond, &params->writer_mutex);
		}

		if (params->exit_thread) {
			process = false;
		}
		has_data_to_write = params->chunk_page_prep != params->chunk_page_write;
		pthread_mutex_unlock(&params->writer_mutex);
		D6(fprintf(debug_file, "_w00.5_: (%d, %d) @ %07d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));
		if (has_data_to_write) {
			D6(fprintf(debug_file, "_w01_: got write data (%d > %d) @ %07d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));
			l = 0;
			reset_rem = 0;
			state->writer_params.last_ret_val = 0;
			 // inputs from state->writer_params.data_chunks, returns number of result chunks
			wpage = state->writer_params.chunk_page_write % SEGMENTS_PAGES;
			num_write_segments = 0;
			for (i = 0; i < SEGMENTS_NUMBER; i++){
				seg_len = state->writer_params.data_segments[wpage][i].iov_len;
				if (seg_len > 0){
					write_segments[num_write_segments++] = state->writer_params.data_segments[wpage][i];
					l+=seg_len;
				}
			}
			// chunk_index = get_data_buffers(state, chunks_iovec[wpage], FILE_CHUNKS_NUM, wpage);
			// assuming write is block-aligned
			if (num_write_segments > 0) {
				// Moved rollover here
				blocks_write = l / PHY_BLOCK_SIZE; // here LBA, 512
				if (state->writer_params.lba_current + blocks_write <= state->writer_params.lba_end) { // almost always - single write
					state->writer_params.lba_current += blocks_write;
					D2(dbg_us1=get_fpga_usec(state->fd_fparmsall[0], 0));
					D6(dbg_pos = lseek(state->writer_params.blockdev_fd, 0, SEEK_CUR));
					D6(fprintf(debug_file, "_w02_: starting emul_writev@ %07d\n", dbg_us1));
					state->writer_params.writev_run = true;
					pthread_cond_signal(&params->main_cond);
					iovlen = emul_writev(state->writer_params.blockdev_fd, write_segments, num_write_segments);
					state->writer_params.writev_run = false;
					D2(dbg_us=get_fpga_usec(state->fd_fparmsall[0], 0); fprintf(debug_file, "_w03_: finished emul_writev, pos%%block = %ld, pos= %ld  @ %07d (+ %07d, +%07d)\n",dbg_pos % PHY_BLOCK_SIZE, dbg_pos,dbg_us, (dbg_us-dbg_us1), (dbg_us-dbg_us2)));
					D3(syslog (LOG_INFO, "_w03_: finished emul_writev, pos= %ld, len = %d(0x%x) @ %07d (+ %07d, +%07d)", dbg_pos, iovlen, iovlen, dbg_us, (dbg_us-dbg_us1), (dbg_us-dbg_us2)));
					D2(dbg_us2=dbg_us);
					if (iovlen < l) {
						D0(fprintf(debug_file, "emul_writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, l));
						state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
						reset_rem = 1;
						lseek(state->writer_params.blockdev_fd,
								(state->writer_params.lba_current - state->writer_params.lba_start) * PHY_BLOCK_SIZE,
								SEEK_SET); // update as if written?
					} else {
						// update statistic
						state->rawdev.last_jpeg_size = l;
						state->rawdev.total_rec_len += state->rawdev.last_jpeg_size;
						D6(fprintf(debug_file, "Current position in block device: %lld\n", lseek64(state->writer_params.blockdev_fd, 0, SEEK_CUR)));
					}
				} else { // Split into two writes: to the end of partition and startion from the beginning
					l0 = (state->writer_params.lba_end - state->writer_params.lba_current) * PHY_BLOCK_SIZE; // remaining to be written before rollover
					l1 = l - l0; // remaining to be written after rollover
					i1 = 0; // start of the second write
					if (l0 > 0) { // should be always
						ll = 0;
						for (i0 = 0; i0 < num_write_segments; i0++){
							if (ll + write_segments[i0].iov_len > l0){
								break;
							}
							ll += write_segments[i0].iov_len;
						}
						// save split chunk
						split_chunk.iov_base = write_segments[i0].iov_base;
						split_chunk.iov_len =  write_segments[i0].iov_len;
						i1 = i0; // index of the first chunk in a second write
						if (ll < l0) { // is it a partial chunk?
							write_segments[i0].iov_len = l0 - ll;
							i0++;     // include split chunk in the first write
						}
						// write the first group (before rollover)
						state->writer_params.lba_current = state->writer_params.lba_start;
						state->writer_params.writev_run = true;
						pthread_cond_signal(&params->main_cond);
						iovlen = emul_writev(state->writer_params.blockdev_fd, write_segments, i0);
						state->writer_params.writev_run = false;
						lseek(state->writer_params.blockdev_fd, 0, SEEK_SET); // regardless of error - rewind
						if (iovlen < l0) {
							D0(fprintf(debug_file, "(1 of 2) emul_writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, l0));
							state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
							reset_rem = 1;
						} else {
							// do not update statistic after the first emul_writev(), do it after the second
							D6(fprintf(debug_file, "(1 of 2) Current position in block device: %lld\n", lseek64(state->writer_params.blockdev_fd, 0, SEEK_CUR)));
							// update split chunk to use its second part, shift chunks
							if (l1 > 0) { // length to be written after rollover
								if (write_segments[i1].iov_len != split_chunk.iov_len) { // it was split
									D0(fprintf(debug_file, "chunk %d was split, full length = %d, writtren = %d\n", i1, split_chunk.iov_len, write_segments[i1].iov_len));
									write_segments[i1].iov_base += write_segments[i1].iov_len;
									write_segments[i1].iov_len = split_chunk.iov_len - write_segments[i1].iov_len;
								}
								// shift chunks
								for (i0 = i1; i0 < num_write_segments; i0++){
									write_segments[i0 - i1].iov_base = write_segments[i0].iov_base;
									write_segments[i0 - i1].iov_len =  write_segments[i0].iov_len;
								}
								state->writer_params.writev_run = true;
								pthread_cond_signal(&params->main_cond);
								iovlen = emul_writev(state->writer_params.blockdev_fd, write_segments, num_write_segments - i1);
								state->writer_params.writev_run = false;

								state->writer_params.lba_current = state->writer_params.lba_start + l1 / PHY_BLOCK_SIZE;
								if (iovlen < l1) {
									D0(fprintf(debug_file, "(2 of 2) emul_writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, l1));
									state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
									reset_rem = 1;
								} else {
									// update statistic (for both parts of the split write)
									state->rawdev.last_jpeg_size = l;
									state->rawdev.total_rec_len += state->rawdev.last_jpeg_size;
									D6(fprintf(debug_file, "(1 of 2) Current position in block device: %lld\n", lseek64(state->writer_params.blockdev_fd, 0, SEEK_CUR)));
								}
							}
						}
					}
				}
				D6(fprintf(debug_file, "_w04: wrote data @ %07d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
				D6(fprintf(debug_file, "_w05_: Block device positions: start = %llu, current = %llu, end = %llu\n",
						state->writer_params.lba_start, state->writer_params.lba_current, state->writer_params.lba_end));
			} else {
				D0(fprintf(debug_file, "data vector mapping error: %d)\n", num_write_segments));
				state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
				reset_rem = 1;
			}
			// release main thread
			reset_segments(state, reset_rem, state->writer_params.chunk_page_write % SEGMENTS_PAGES);
			pthread_mutex_lock(&params->writer_mutex);
			params->chunk_page_write = (params->chunk_page_write + 1) % (2 * SEGMENTS_PAGES);
			if (params->chunk_page_prep == params->chunk_page_write) {
				pthread_cond_signal(&params->main_cond); // **********  Do not signal if can write more !
				D6(fprintf(debug_file, "_w06_: wrote data (%d , %d), buffer empty, signal main thread @ %07d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));
			} else {
				D6(fprintf(debug_file, "_w06a_: wrote data (%d , %d), has more to write, do not signal! @ %07d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));

			}
			pthread_mutex_unlock(&params->writer_mutex); // do not release untill waiting for emul_writev() or waiting for data to write
		}
	}
	params->state = STATE_STOPPED;
	D5(fprintf(debug_file, "_w07_: Exit from recording thread\n")); // never?
	return NULL;
}
